{*****************************************************************************}
{                                                                             }
{                           Objects on the Stack                              }
{                                                                             }
{               Copyright (c) 2001, 2005 Robert R. Marsh, S.J.                }
{               & the British Province of the Society of Jesus                }
{                                                                             }
{                                                                             }
{       If you like this code and find yourself using it then please          }
{       consider making a donation to your favorite charity. I would          }
{       also be pleased if you would acknowledge me in any projects           }
{       that make use of them.                                                }
{                                                                             }
{       This unit is supplied as is. The author disclaims all warranties,     }
{       expressed or implied, including, without limitation, the warranties   }
{       of merchantability and of fitness for any purpose.                    }
{                                                                             }
{       The author assumes no liability for damages, direct or                }
{       consequential, which may result from use of this code.                }
{                                                                             }
{                          stack@rmarsh.com                                   }
{                        http://www.rmarsh.com                                }
{                                                                             }
{*****************************************************************************}

(*

   Very fast allocation of objects (and other creatures) on the stack

*)

unit StackObjects;

interface

uses
  SysUtils;

type
  EStackMemOverflow = class(Exception);

type
  TStackStatus = record
    Capacity: Cardinal; // total size of reserved space
    Available: Cardinal; // amount of space still available
  end;

function StackStatus: TStackStatus;

// Initialize the stack memory manager for the current
// procedure. <size> bytes will be available in all.
// Exceeding this limit will raise EStackError.
// String allocations are always passed along to the standard
// memory manager. Depending on the value of ObjectsOnly all other
// allocations or just objects will be made on the stack.
procedure BeginStackAllocation(Size: Cardinal; ObjectsOnly: Boolean);

// Restore the default memory manager.
procedure EndStackAllocation;

implementation

var
  OldMM: TMemoryManager;
  IsNewMM: Boolean = False;

var
  // global pointer to the reserved stack space
  StackMemory: Pointer = nil;
  // offset from the start of StackMemory
  StackOffset: Cardinal = 0;
  // size of StackMemory block
  StackCapacity: Cardinal = 0;
  // ObjOnly is true by default and only TObject
  // descendants are allocated on the stack. Other
  // allocations (e.g., ordinary memory, dynamic arrays)
  // remain on the heap. AnsiStrings and WideStrings are
  // never shifted to the stack since they cause a lot
  // of problems.
  // If ObjOnly is set to false you should be VERY
  // careful to make sure reference-counted variables
  // like dynamic arrays are finalized before the
  // standard memory manager is reinstated with EndStackAllocation.
  ObjOnly: LongBool = True;

var
  // string [de]allocations always come between Lo- and HiStringAdr
  LoStringAdr: Cardinal;
  HiStringAdr: Cardinal;
  // object [de]allocations always come between Lo- and HiObjectAdr
  LoObjectAdr: Cardinal;
  HiObjectAdr: Cardinal;

// Returns a pointer into an unused portion of the reserved stack,
// effectively allocating memory.

function StackGetMem(Size: Integer): Pointer;
var
  Caller: Cardinal;
begin
  // get the address of the calling routine
  asm
    mov   eax,[ebp+8]
    mov   Caller,eax
  end;
  // if the caller is a string operation it is always excluded
  if ((Caller > LoStringAdr) and (Caller < HiStringAdr))
  // and if ObjectOnly is true only objects get through
  or (ObjOnly and ((Caller < LoObjectAdr) or (Caller > HiObjectAdr))) then
  begin
    Result := OldMM.GetMem(Size);
  end
  else
  begin
    // return a pointer into the reserved stack
    Result := Pointer(Cardinal(StackMemory) + StackOffset);
    Inc(StackOffset, Size);
    if StackOffset > StackCapacity then
    begin
      EndStackAllocation;
      raise EStackMemOverflow.Create('stack capacity exceeded');
    end;
  end;
end;

// Deallocates previously allocated memory. In fact, stack
// memory is not freed since it automatically disappears
// when pointer goes out of scope.

function StackFreeMem(P: Pointer): Integer;
var
  Caller: Cardinal;
begin
  // get the address of the calling routine
  asm
    mov   eax,[ebp+8]
    mov   Caller,eax
  end;
  // if the caller is a string operation it is always excluded
  if ((Caller > LoStringAdr) and (Caller < HiStringAdr))
  // and if ObjectOnly is true only objects get through
  or (ObjOnly and ((Caller < LoObjectAdr) or (Caller > HiObjectAdr))) then
  begin
    Result := OldMM.FreeMem(P);
  end
  else
  begin
    // do nothing except report success
    Result := 0;
  end;
end;

// Reallocation is achieved by returning a fresh pointer.

function StackReallocMem(P: Pointer; Size: Integer): Pointer;
var
  Caller: Cardinal;
begin
  // get the address of the calling routine
  asm
    mov   eax,[ebp+12]   // note 12 and not 8 since we have two parameters
    mov   Caller,eax
  end;
  // if the caller is a string operation it is always excluded
  if ((Caller > LoStringAdr) and (Caller < HiStringAdr))
  // and if ObjectOnly is true only objects get through
  or (ObjOnly and ((Caller < LoObjectAdr) or (Caller > HiObjectAdr))) then
  begin
    Result := OldMM.ReallocMem(P, Size);
  end
  else
  begin
    // we don't really reallocate ... we just get a
    // fresh pointer into the reserved stack ...
    Result := Pointer(Cardinal(StackMemory) + StackOffset);
    Inc(StackOffset, Size);
    if StackOffset > StackCapacity then
    begin
      EndStackAllocation;
      raise EStackMemOverflow.Create('stack capacity exceeded');
    end;
    // ... and copy the contents of old to new ...
    // even though the size parameter is wrong (we don't
    // know the 'old' size) we can use it since Realloc
    // guarantees nothing about the contents any 'new' memory
    Move(P^, Result^, Size);
  end;
end;

var
  NewMM: TMemoryManager = (GetMem: StackGetMem; FreeMem: StackFreeMem; ReallocMem: StackReallocMem);

// just swap the memory managers

procedure SetStackMemoryManager;
begin
  Assert(not IsNewMM, 'BeginStackAllocation cannot be nested');
  GetMemoryManager(OldMM);
  SetMemoryManager(NewMM);
  IsNewMM := True;
  StackOffset := 0;
end;

// based on StackAlloc in grids.pas -- thank you Borland!

procedure BeginStackAllocation(Size: Cardinal; ObjectsOnly: Boolean); register;
asm
  POP   ECX                    // return address
  MOV   ObjOnly, EDX           // which allocations should use the stack?
  MOV   EDX, ESP
  ADD   EAX, 3
  AND   EAX, not 3             // round up to keep ESP dword aligned
  MOV   StackCapacity, EAX     // remember the size
  CMP   EAX, 4092
  JLE   @@2
@@1:
  SUB   ESP, 4092
  PUSH  EAX                    // make sure we touch guard page, to grow stack
  SUB   EAX, 4096
  JNS   @@1
  ADD   EAX, 4096
@@2:
  SUB   ESP, EAX
  MOV   EAX, ESP               // function result = low memory address of block
  MOV   StackMemory, EAX       // set the global pointer
  PUSH  EDX                    // save original SP, for cleanup
  MOV   EDX, ESP
  SUB   EDX, 4
  PUSH  EDX                    // save current SP, for sanity check  (sp = [sp])
  PUSH  ECX                    // return to caller
  CALL  SetStackMemoryManager  // swap the managers
end;

procedure EndStackAllocation;
begin
  Assert(IsNewMM, 'StackAllocation is not in action');
  SetMemoryManager(OldMM);
  IsNewMM := False;
end;

function StackStatus: TStackStatus;
begin
  Result.Capacity := StackCapacity;
  Result.Available := StackCapacity - StackOffset;
end;

initialization
  // these values work for D3-D5 but may need changing in later versions... 
  // they are the first routine before and after the routines we want to trap,
  // i.e., all string allocation stuff comes somewhere between
  // EndThread and WideCharToString in system.pas ...
  LoStringAdr := Cardinal(@EndThread);
  HiStringAdr := Cardinal(@WideCharToString);
  LoObjectAdr := Cardinal(@TObject.ClassType);
  HiObjectAdr := Cardinal(@TObject.InstanceSize);
end.

